unit SimpsonMethod;

////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                ()              //
//                                                   //
//                                                                            //
//                             SIMPSON v.1.55                                 //
//                                                                            //
//  ..                                                               //
//  2011                                                              //
// sadovoya@mtu-net.ru                                                        //
// sadovoya.narod.ru                                                          //
//                                                                            //
//     .  . " ".                  //
////////////////////////////////////////////////////////////////////////////////


//------------------------------------------------------------------------------
//
//   :
//
// I = (x2 - x0)*(f(x0) + 4*f(x1) + f(x2))*(1/6)
// : x0 -   
//      x2 -  
//      x1 -  
//      I -     f   [x0; x2]
//          (I -      0 - 3 )
//
//    ,      ,
//    ,  :
//
// I = (h/6)*(f_. + f_. + 4*Sum_. + 2*Sum)
// : h -   (  3   )
//      f_.  f_. -       
//                          
//      Sum_. -      
//      Sum -        ( f_.
//             f_.)
//
//------------------------------------------------------------------------------


interface

type
  TFunOfX_Extended = function(const x: Extended): Extended;
  TFunOfX_Double   = procedure(const x: Double; out f: Double);

function Simpson(fun: TFunOfX_Extended;
                 x0, xk: Extended;
                 ParabolasNum: Integer): Extended; overload;

function Simpson(fun: TFunOfX_Extended;
                 x0, xk, Prec: Extended): Extended; overload;



function Simpson(fun: TFunOfX_Double;
                 x0, xk: Double;
                 ParabolasNum: Integer): Double; overload;

function Simpson(fun: TFunOfX_Double;
                 x0, xk, Prec: Double): Double; overload;

implementation

uses SysUtils;

//------------------------------------------------------------------------------

// .

function Sum(fun: TFunOfX_Extended; FromX, dx: Extended;
             Num: Integer):Extended; overload;
var
   I : Integer;
begin
   Result := 0;
   for I := 0 to Num do Result := Result + fun(FromX + I*dx);

end;

function Sum(fun: TFunOfX_Double; FromX, dx: Double;
             Num: Integer):Double;  overload;
var
   I : Integer;
   f : Double;
begin
   Result := 0;
   for I := 0 to Num do
   begin
     fun(FromX + I*dx, f);
     Result := Result + f;
   end;
end;

//------------------------------------------------------------------------------

//  Extended:

function Simpson(fun: TFunOfX_Extended;
                 x0, xk: Extended; ParabolasNum: Integer): Extended;
//   ,     ,   ,
//    -    .
// Extended ,     ,  
// Sum().
var
   x, f1, f3, dx, half_dx: Extended;
   I : Integer;
begin
  If ((x0 >= xk) or (ParabolasNum <= 0)) then
    Raise Exception.Create('Simpson ERROR - BAD_ARGUMENTS');
  try
    Result := 0;
    dx := (xk - x0)/ParabolasNum;
    half_dx := dx/2.0;
    f1 := fun(x0); //. f1
    for I := 0 to ParabolasNum - 1 do
    begin
      x := x0 + I*dx; //  . ,   dx+dx+...
      f3 := fun(x + dx);
      Result := Result + f1 + 4.0*fun(x + half_dx) + f3;
      f1 := f3; //    =   
    end;
    Result := dx*Result/6.0;
  except
    Raise Exception.Create('Simpson ERROR - INTEGRAT_ERROR');
  end;
end;

{function Simpson(fun: TFunOfX_Extended; x0, xk, Prec: Extended): Extended;
var
   x, dx, half_dx, Sum1, Sum2, Sum_const, Delt, Int2: Extended;
   I, Steps : Integer;
begin
  If Prec <= 0 then
       Raise Exception.Create('Simpson ERROR - BAD_PREC');
  try
    Sum_const := fun(x0) + fun(xk);
    Steps := 10;
    dx := (xk - x0)/Steps;
    half_dx := dx/2.0;
    Sum1 := 0;
    Sum2 := 0;
    for I := 0 to 2*Steps do  //  . ,   dx+dx+...
          if (I mod 2 = 0) then Sum2 := Sum2 + fun(x0 + I*half_dx)
          else Sum1 := Sum1 + fun(x0 + I*half_dx);
    Result := dx*(4.0*Sum1 + 2.0*Sum2 - Sum_const )/6.0;
    repeat
        Steps := 2*Steps; // 2
        dx := (xk - x0)/Steps;
        Sum2 := Sum1 + Sum2;
        Sum1 := 0;
        x := x0 + dx/2.0;
        for I := 0 to Steps - 1 do  Sum1 := Sum1 + fun(x + I*dx);
        Int2 := dx*(4.0*Sum1 + 2.0*Sum2 - Sum_const)/6.0;
        Delt := Int2 - Result;
        if Delt < 0 then Delt := -Delt; //abs()   
        Result := Int2;
    until Delt < Prec;
  except
    Raise Exception.Create('Simpson ERROR - INTEGRAT_ERROR');
  end;
end; }

function Simpson(fun: TFunOfX_Extended; x0, xk, Prec: Extended): Extended;
//    Sum(). ,      
//        .  
//       .
var
    Int2, Delt, dx, Sum1, Sum2, Sum_const: Extended;
    Steps: Integer;
begin
    If Prec <= 0 then
       Raise Exception.Create('Simpson ERROR - BAD_PREC');
    try
      Steps := 10;
      dx := (xk - x0)/Steps;
      Sum1 := Sum(fun, x0 + dx/2.0, dx, Steps - 1);
      Sum2 := Sum(fun, x0 + dx, dx, Steps - 2);
      Sum_const := fun(x0) + fun(xk); //  
      Result := (dx/6.0)*(2.0*Sum2 + 4.0*Sum1 + Sum_const);
      repeat
        Steps := 2*Steps; // 2
        Sum2 := Sum2 + Sum1; // 
        dx := (xk - x0)/Steps;
        Sum1 := Sum(fun, x0 + dx/2.0, dx, Steps - 1);
        Int2 := (dx/6.0)*(2.0*Sum2 + 4.0*Sum1 + Sum_const);
        Delt := Int2 - Result;
        if Delt < 0 then Delt := -Delt; //abs()   
        Result := Int2;
      until Delt < Prec;
    except
      Raise Exception.Create('Simpson ERROR - INTEGRAT_ERROR');
    end;
end;
             

//  Double:

function Simpson(fun: TFunOfX_Double;
                 x0, xk: Double; ParabolasNum: Integer): Double;
var
   f, dx: Double;
begin
  If ((x0 >= xk) or (ParabolasNum <= 0)) then
    Raise Exception.Create('Simpson ERROR - BAD_ARGUMENTS');
  try
    dx := (xk - x0)/ParabolasNum;
    fun(x0, f);
    Result := f;
    fun(xk, f);
    Result := (dx/6.0)*(2.0*Sum(fun, x0 + dx, dx, ParabolasNum - 2)
       + 4.0*Sum(fun, x0 + dx/2.0, dx, ParabolasNum - 1)
       + f + Result);
  except
    Raise Exception.Create('Simpson ERROR - INTEGRAT_ERROR');
  end;
end;

function Simpson(fun: TFunOfX_Double; x0, xk, Prec: Double): Double;
var
    Int2, Delt, dx, Sum1, Sum2, Sum_const: Double;
    Steps: Integer;
begin
    If Prec <= 0 then
       Raise Exception.Create('Simpson ERROR - BAD_PREC');
    try
      Steps := 10;
      dx := (xk - x0)/Steps;
      Sum1 := Sum(fun, x0 + dx/2.0, dx, Steps - 1);
      Sum2 := Sum(fun, x0 + dx, dx, Steps - 2);
      fun(x0, Result);
      fun(xk, Sum_const);
      Sum_const := Sum_const + Result;
      Result := (dx/6.0)*(2.0*Sum2 + 4.0*Sum1 + Sum_const);
      repeat
        Steps := 2*Steps;
        Sum2 := Sum2 + Sum1;
        dx := (xk - x0)/Steps;
        Sum1 := Sum(fun, x0 + dx/2.0, dx, Steps - 1);
        Int2 := (dx/6.0)*(2.0*Sum2 + 4.0*Sum1 + Sum_const);
        Delt := Int2 - Result;
        if Delt < 0 then Delt := -Delt;
        Result := Int2;
      until Delt < Prec;
    except
      Raise Exception.Create('Simpson ERROR - INTEGRAT_ERROR');
    end;
end;


end.
